import importlib
import traceback
from inspect import getfullargspec
from os import environ
from pathlib import Path
from sys import stderr
from typing import Any, Callable, Dict, List, Optional

from entest.const import STATUS, display
from entest.status_report import format_error

PRINT_TO_STDERR = lambda *args, **kwargs: print(*args, **kwargs, file=stderr)


class TestCase:
    full_registry: Dict[str, "TestCase"] = {}
    snoop_decorator: Callable[[Callable], Callable] = lambda self, x: x
    dynamic_globals: Dict[str, Any] = {}
    children: List["TestCase"]
    without: Optional["TestCase"]

    def __init__(self, func: Callable[[], None], parents: List["TestCase"], run_last=False):
        for parent in parents:
            #  if not isinstance(parent, TestCase):
            #      raise Exception("cannot depend on undecorated function")
            parent.children.append(self)
        self.func = func
        self.parents = list(parents)
        self.run_last = run_last
        self.children = []
        self.without = None
        self.status = STATUS.none
        if self.func.__name__ in self.full_registry:
            PRINT_TO_STDERR("warning: a test with this name already exists")
        self.full_registry[self.func.__name__] = self
        if parents:
            self.depth: int = max((parent.depth for parent in parents)) + 1
        else:
            self.depth = 0
        self.error: Optional[Exception] = None

    def __call__(self) -> bool:
        if self.status == STATUS.none:
            raise Exception("Test is was not expected to run")
        if self.status != STATUS.wait:
            raise Exception("Test is being run more than once")
        if any((test.status in [STATUS.error, STATUS.deps_failed] for test in self.parents)):
            self.status = STATUS.deps_failed
            return False
        if any((test.status not in [STATUS.passed, STATUS.wip] for test in self.parents)):
            raise Exception(
                "Test is being run before its dependencies have finished running.",
                self,
                self.parents,
            )
        if self.without and self.without.status not in [STATUS.none, STATUS.wait]:
            raise Exception(
                f"Test ({self.without}) is not allowed to be run before this test.",
                "See 'without' flag for more info.",
                self,
            )
        self.status = STATUS.running
        required_globals = getfullargspec(self.func).args
        missing_globals = set(required_globals) - set(self.dynamic_globals.keys())
        if missing_globals:
            raise Exception("Cannot run test. Some globals are missing", missing_globals)
        try:
            new_globals = self.snoop_decorator(self.func)(
                *[self.dynamic_globals[key] for key in required_globals]
            )
        except Exception as err:
            self.status = STATUS.error
            self.error = err
            return False
        if new_globals:
            if new_globals == "wip":
                self.status = STATUS.wip
                return False
            if isinstance(new_globals, tuple) and new_globals[0] == "wip":
                self.status = STATUS.wip
                new_globals = new_globals[1]
            if not isinstance(new_globals, dict):
                raise Exception("return value from test case should be a dict")
            self.dynamic_globals.update(new_globals)
        self.status = STATUS.passed
        return True

    def name(self):
        return self.func.__name__

    def display(self):
        return f"{display(self.status):<11}|{' ' * self.depth}{self.name()}"

    def __str__(self):
        return self.name()

    def __repr__(self):
        return self.display()

    @classmethod
    def set_globals_dict(cls, d: dict):
        cls.dynamic_globals = d

    @classmethod
    def summary(cls, as_dict=False):
        report = {
            "total": len(cls.full_registry),
        }
        report.update(
            {
                status: len([None for t in cls.full_registry.values() if t.status == status])
                for status in STATUS
            }
        )
        if as_dict:
            return report
        else:
            return [
                f"""Total: {report['total']}""",
                f"""{display(STATUS.passed)}: {report[STATUS.passed]}""",
                f"""{display(STATUS.error)}: {report[STATUS.error]}""",
                f"""{display(STATUS.deps_failed)}: {report[STATUS.deps_failed]}""",
            ]

    @classmethod
    def print_error_summary(cls, logger) -> None:
        for test in cls.full_registry.values():
            if test.error is None:
                continue
            logger(f"""===== {test.name()} =====""")
            logger(format_error(test.error))
            logger()


def empty_setup():
    pass


TEST_ROOT = TestCase(empty_setup, [])  ## make this singleton


def setup_setup(callback: Callable[[], None]) -> None:
    TEST_ROOT.func = callback


class DependsOn:
    def __init__(self):
        self.fixtures = {}
        self.last_file_handled = ""
        self.last_decorated = TEST_ROOT

    def __call__(
        self, *requires: TestCase, previous: bool = None, without: TestCase = None, run_last=False
    ) -> Callable[[Callable], TestCase]:
        """
        For "previous" keyword to work as expected this decorator should be the last one applied.
        "run_last" is a temporary workaround to mark tests that delete important resources.
        "without" is used to mark a single test as a must NOT have dependency
        """
        if previous is None:
            previous = len(requires) == 0

        # -1 is the current file "dependency_decorator.py" and -2 is the one we want
        current_file = traceback.extract_stack()[-2].filename
        if current_file != self.last_file_handled:
            self.last_file_handled = current_file
            self.last_decorated = TEST_ROOT

        def decorator(func):
            parents = list(requires)
            if previous:
                parents.append(self.last_decorated)
            test_case = TestCase(func, parents, run_last=run_last)
            if without:
                test_case.without = without
            self.last_decorated = test_case
            return test_case

        return decorator


depends_on = DependsOn()


def remove_implicit_edges(dfs_path: List[TestCase], logger=PRINT_TO_STDERR):
    head = dfs_path[-1]
    for child in head.children:
        for parent in dfs_path[:-1]:
            if child in parent.children:
                logger("Warning: Transitive node detected. From", parent.name(), "to", child.name())
                parent.children.remove(child)
                child.parents.remove(parent)
        dfs_path.append(child)
        remove_implicit_edges(dfs_path)
    dfs_path.pop()


def propogate_status_none() -> int:
    unvisited = {tc for tc in TestCase.full_registry.values() if tc.status == STATUS.wait}
    number_of_tests_to_run = len(unvisited)
    while unvisited:
        new_unvisited = set()
        for test_case in unvisited:
            number_of_tests_to_run += int(test_case.status != STATUS.wait)
            test_case.status = STATUS.wait
            new_unvisited.update(test_case.parents)
        unvisited = new_unvisited
    return number_of_tests_to_run


def test_discovery(dirs: List[Path], logger=PRINT_TO_STDERR):
    logger('Collecting tests')
    if TEST_ROOT.children:
        raise Exception("do not run test_discovery twice")
    paths = list(Path('tests').glob("**/*.py"))
    shortlist: List[Path] = []
    if dirs:
        for dir in dirs:
            shortlist.extend(dir.glob("**/*.py") if dir.is_dir() else [dir])
    else:
        shortlist = paths
    for path in paths:
        file_contents = path.read_bytes()
        if b"from entest" in file_contents or b"import entest" in file_contents:
            lib = importlib.import_module(str(path).replace("/", ".")[:-3])
            if path in shortlist:
                for var in lib.__dict__.values():
                    if isinstance(var, TestCase):
                        var.status = STATUS.wait
                        if environ.get("ENTEST_SKIP_TEARDOWN", False) and var.run_last:
                            var.status = STATUS.none
    logger(f'Collected {len(TestCase.full_registry)} tests')
    logger('Removing implicit edges')
    remove_implicit_edges([TEST_ROOT], logger)
    logger('Resolving dependencies')
    number_of_tests_to_run = propogate_status_none()
    logger(f'Will run {number_of_tests_to_run} tests')



import random
import threading
from collections import defaultdict
from os import abort
from typing import Callable

from entest.const import STATUS
from entest.dependency_decorator import TEST_ROOT, TestCase
from entest.status_report import IS_STACKPRINTER, StatusPanel


def breadth_first_traverse(root: TestCase = TEST_ROOT):
    order = []
    nodes = [root]
    current_depth = 0
    while nodes:
        order.extend(nodes)
        current_depth += 1
        nodes = [test for test in TestCase.full_registry.values() if test.depth == current_depth]
    order.sort(key=lambda test_case: int(test_case.run_last))
    return order


def bogo_order(seed: int = 0):
    random.seed(seed)
    order = []
    unvisited = [test for test in TestCase.full_registry.values()]
    danger_zone = defaultdict(int)
    for test in filter(lambda test: test.without, unvisited):
        danger_zone[test.without] += 1
    while unvisited:
        random.shuffle(unvisited)
        start_len = len(unvisited)
        for test in unvisited:
            if all(parent in order for parent in test.parents) and danger_zone[test] <= 0:
                order.append(test)
                unvisited.remove(test)
                danger_zone[test.without] -= 1
        if len(unvisited) == start_len:
            raise Exception(
                "Impossible to traverse all graph nodes. Use --graph to review dependencies",
            )
    order.sort(key=lambda test_case: int(test_case.run_last))
    return order


def generate_run_sequence(*, strategy=bogo_order):
    order = strategy()
    return [tc for tc in order if tc.status == STATUS.wait]


def run_tests(logger: Callable[..., None]):
    logger("""=============== Running tests ===============""")
    test_order = generate_run_sequence()
    with StatusPanel(test_order, logger):
        for test in test_order:
            test()
    logger("""=================== ERRORS ==================""")
    # It is an ugly hack to use IS_STACKPRINTER here but logging should be refactored in general
    TestCase.print_error_summary(logger if not IS_STACKPRINTER else print)
    join_dangling_threads(logger)
    logger("""================== Summary ==================""")
    logger("\n".join(TestCase.summary()))


def join_dangling_threads(logger: Callable[..., None]) -> None:
    thread_count = threading.active_count() - 1
    if thread_count == 0:
        return
    logger(f"Found {thread_count} dangling threads.")
    for thread in threading.enumerate():
        if thread is threading.current_thread():
            continue
        else:
            thread.join(timeout=10)
            if thread.is_alive:
                logger("Some non-daemon threads could not be joined.", flush=True)
                abort()



__all__ = ["depends_on", "setup_setup", "STATUS", "assert_raises"]
from typing import Callable

from .const import STATUS
from .dependency_decorator import depends_on, setup_setup


def assert_raises(cb: Callable, message="") -> Exception:
    try:
        cb()
    except Exception as err:
        assert message in str(err)
        return err
    raise Exception("Function did not raise Exception")



from typing import NamedTuple, NewType

StatusType = NewType("StatusType", str)


class Status(NamedTuple):
    wip: StatusType = StatusType("wip")
    none: StatusType = StatusType("none")
    wait: StatusType = StatusType("wait")
    running: StatusType = StatusType("running")
    deps_failed: StatusType = StatusType("deps_failed")
    passed: StatusType = StatusType("passed")
    error: StatusType = StatusType("error")


STATUS = Status()


def display(s: str) -> str:
    if s == STATUS.none:
        return "🤷 Ignored"
    elif s == STATUS.wait:
        return "⌛ Not run"
    elif s == STATUS.running:
        return "🏃 Running"
    elif s == STATUS.wip:
        return "🚧 WIP    "
    elif s == STATUS.deps_failed:
        return "⛔ Skipped"
    elif s == STATUS.passed:
        return "✅ Passed "
    elif s == STATUS.error:
        return "❌ Error  "
    else:
        return s



import threading
import time
from sys import stderr
from traceback import format_tb

try:
    from rich.console import Console
    from rich.live import Live
    from rich.table import Table
    from rich.traceback import Traceback

    IS_RICH = True
except ImportError:
    IS_RICH = False
    Table = None

try:
    import stackprinter

    IS_STACKPRINTER = True
except ImportError:
    stackprinter = None
    IS_STACKPRINTER = False
from entest.const import STATUS, display

if IS_RICH:
    console = Console()
    stderr_console = Console(stderr=True)

    def logger(*args) -> None:
        console.print(*args)

    def stderr_logger(*args) -> None:
        stderr_console.print(*args)

else:

    def logger(*args) -> None:
        print(" ".join(args))

    def stderr_logger(*args) -> None:
        print(*args, file=stderr)


def generate_table(order: list) -> Table:
    """Make a new table."""
    table = Table()
    table.add_column("Status")
    table.add_column("Test Name")
    for test in order:
        table.add_row(display(test.status), f"{' ' * test.depth}{test.name()}")
    return table


def poor_panel(order, logger):
    pending_tests = order.copy()
    while any(test.status in (STATUS.wait, STATUS.running) for test in order):
        just_finished_tests = filter(
            lambda test: test.status not in (STATUS.wait, STATUS.running),
            pending_tests,
        )
        for test in just_finished_tests:
            logger(test.display())
            pending_tests.remove(test)
        time.sleep(0.4)


def rich_panel(order):
    with Live(generate_table(order), refresh_per_second=4) as live:
        while any(test.status in (STATUS.wait, STATUS.running) for test in order):
            time.sleep(0.4)
            live.update(generate_table(order))


class StatusPanel:
    def __init__(self, order, logger):
        if IS_RICH:
            self.thread = threading.Thread(target=rich_panel, args=(order,))
        else:
            self.thread = threading.Thread(target=poor_panel, args=(order, logger))

    def __enter__(self):
        self.thread.start()

    def __exit__(self, *args):
        self.thread.join(timeout=1)


def format_error(error: Exception):
    if IS_STACKPRINTER:
        return stackprinter.format(error, style="darkbg3")
    if IS_RICH:
        return Traceback(Traceback.extract(type(error), error, error.__traceback__))
    else:
        return type(error).__name__ + "\n" + "".join(format_tb(error.__traceback__))



#!/usr/bin/env python3
import argparse
import os
import sys
from pathlib import Path

from entest.const import STATUS
from entest.dependency_decorator import TestCase, test_discovery
from entest.graph import graph
from entest.runner import run_tests
from entest.snoop import setup_snooper
from entest.status_report import logger, stderr_logger

parser = argparse.ArgumentParser(description='Run integration tests.')
parser.add_argument('paths', type=str, nargs="*", help='files to run')
parser.add_argument('--graph', action="store_true", help="Print Mermaid diagram of the test tree")
parser.add_argument(
    '--snoop',
    default="",
    const="std",
    nargs="?",
    help="Print automatic debugging to a file or terminal if no file is provided",
)
parser.add_argument(
    '--env', default="", help="Sets TEST_ENV_NAME env variable used by env/loader.py"
)
parser.add_argument(
    '--skip-teardown', action="store_true", help="Do not run test cases that have `run_last=True`"
)


def main():
    args = parser.parse_args()
    if args.env:
        os.environ["TEST_ENV_NAME"] = args.env
    if args.skip_teardown:
        os.environ["ENTEST_SKIP_TEARDOWN"] = "yes"
    if args.snoop:
        setup_snooper(args.snoop)
    paths = [Path(path) for path in args.paths]
    test_discovery(paths, stderr_logger)
    if args.graph:
        logger(graph())
        sys.exit(0)
    run_tests(logger)
    sys.exit(TestCase.summary(as_dict=True)[STATUS.error])


if __name__ == '__main__':
    main()



from typing import Callable, Optional, TypeVar

from entest.dependency_decorator import TestCase

F = TypeVar("F", bound=Callable)


def setup_snooper(output: Optional[str]) -> Callable[[F], F]:
    try:
        from pysnooper import snoop
    except ImportError as err:
        raise ImportError("You need to install optional dependency 'pysnooper' to snoop") from err

    if output == "std":
        TestCase.snoop_decorator = snoop()
    else:
        TestCase.snoop_decorator = snoop(output)



from io import StringIO
from typing import List

from entest.const import STATUS
from entest.dependency_decorator import TEST_ROOT, TestCase


def mermaid_edges(node: TestCase):
    node_name = node.name()
    test_edges = []
    teardown_edges = []
    for child in node.children:
        if child.run_last:
            teardown_edges.append(f"{node_name} --> {child.name()}")
        else:
            test_edges.append(f"{node_name} --> {child.name()}")
    if node.without:
        if node.run_last:
            teardown_edges.append(f"{node.without.name()} -.- {node_name}")
        else:
            test_edges.append(f"{node.without.name()} -.- {node_name}")
    return test_edges, teardown_edges


output = StringIO()


def printer(s, prefix="  "):
    print(prefix, s, file=output)


def graph(root=TEST_ROOT):
    eligible_tests = filter_nodes(root.children)
    if len(eligible_tests) == 0:
        return ""
    unvisited_nodes = set(eligible_tests)
    test_edges = []
    teardown_edges = []
    visited_nodes = {root}
    while unvisited_nodes:
        node = unvisited_nodes.pop()
        visited_nodes.add(node)
        unvisited_nodes.update([child for child in node.children if child not in visited_nodes])
        new_test_edges, new_teardown_edges = mermaid_edges(node)
        test_edges.extend(new_test_edges)
        teardown_edges.extend(new_teardown_edges)
    printer("flowchart TD", prefix="")
    printer("subgraph tests")
    for edge in test_edges:
        printer(edge)
    printer("end")
    if len(teardown_edges) > 0:
        printer("subgraph teardown")
        for edge in teardown_edges:
            printer(edge)
        printer("end")
    return output.getvalue()


def filter_nodes(nodes: List["TestCase"]):
    return [node for node in nodes if node.status != STATUS.none]
